/* 
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File vgaspectrum.c - alternative version of the keypress/release 
 *                      code for when the SAM is in spectrum mode
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>

#include <stdio.h>
#include <stdlib.h>

#include "z80.h"
#include "xspectrum.h"
#include "sam.h"
#include "resource.h"
#include "util.h"

static struct ktbl {
	int port;	/* which keyboard row is the key on? */
	int bit;	/* which bit to mask */
} keytable[128] = {
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, 
	/* SPACE ! "" # $ % & ' */
	{7, 0xfe}, {3, 0xfe}, {9, 0}, {9, 0}, {3,0xf7},
	{3,0xef}, {4,0xf7}, {9, 0},
	/* ( ) * + , - . / */
	{4, 0xfb}, {4, 0xfd}, {4, 0xbf}, {9, 0}, {9, 0}, {9, 0},
	{9, 0}, {9, 0}, 
	/* 0 1 2 3 4 5 6 7 8 9 */
	{4, 0xfe}, {3, 0xfe}, {3, 0xfd}, {3, 0xfb}, {3, 0xf7},
	{3, 0xef}, {4, 0xef}, {4, 0xf7}, {4, 0xfb}, {4, 0xfd},
	/* : ; < = > ? @ */
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}, {3, 0xfd}, 
	/* A -> Z  (upper & lower case are the same) */
	{1, 0xfe}, {7, 0xef}, {0, 0xf7}, {1, 0xfb}, {2, 0xfb}, {1, 0xf7},
	{1, 0xef}, {6, 0xef}, {5, 0xfb}, {6, 0xf7}, {6, 0xfb}, {6, 0xfd}, 
	{7, 0xfb}, {7, 0xf7}, {5, 0xfd}, {5, 0xfe}, {2, 0xfe}, {2, 0xf7},
	{1, 0xfd}, {2, 0xef}, {5, 0xf7}, {0, 0xef}, {2, 0xfd}, {0, 0xfb},
	{5, 0xef}, {0, 0xfd},
	/* [ \ ] ^ _ ` */
	{9, 0}, {9, 0}, {9, 0}, {4, 0xef}, {9, 0}, {9, 0},
	/* a -> z  (upper & lower case are the same) */
	{1, 0xfe}, {7, 0xef}, {0, 0xf7}, {1, 0xfb}, {2, 0xfb}, {1, 0xf7},
	{1, 0xef}, {6, 0xef}, {5, 0xfb}, {6, 0xf7}, {6, 0xfb}, {6, 0xfd}, 
	{7, 0xfb}, {7, 0xf7}, {5, 0xfd}, {5, 0xfe}, {2, 0xfe}, {2, 0xf7},
	{1, 0xfd}, {2, 0xef}, {5, 0xf7}, {0, 0xef}, {2, 0xfd}, {0, 0xfb},
	{5, 0xef}, {0, 0xfd},
	/* { | } ~ DEL */
	{9, 0}, {9, 0}, {9, 0}, {9, 0}, {9, 0}
};

/* from xevents.c */
extern int SpectrumKeyboard;
extern uns8 keyports[]; 
extern KeySym *key_table;
extern uns8 inPorts[256], outPorts[256];

/* from processor.c */
extern int disass;
extern int InterruptType;

/* Process a KeyPress event */
void
process_spectrum_keypress(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;
	char *str;
	int i;	


	XLookupString(kev, buf, 2, &ks, NULL);

	/* First check for the 'special keys' */
	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] |= 0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] |= 0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] |= 0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] |= 0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] |= 0x01;
#ifdef SLOWDOWN
	else if (ks == key_table[KM_KEY_SLOWER])
		adjust_speed(1);
	else if (ks == key_table[KM_KEY_FASTER])
		adjust_speed(-1);
#endif

	switch (ks) {
		/* Emulator control keys */
		case XK_F1:
			InterruptType = Z80_reset;
			for (i = 0; i < 9; i++)
				keyports[i] = 0xff;
			break;

		case XK_F2:	/* Switch keyboard routines */
			SpectrumKeyboard = !SpectrumKeyboard;
			break;		
		case XK_F10: case XK_F12:
			Quit(0);
			break;

		/* Ordinary SAM keys */
		case XK_Tab:
			keyports[3] &= 0xbf; break;
		case XK_Return:
			keyports[6] &= 0xfe; break;
		case XK_Control_L: case XK_Control_R:
		/* <Shift> <Control> is mapped to Caps Lock */
			if (!(keyports[0] & 0x01))
			     keyports[3] &= 0xfd;
			break;
		case XK_Shift_L: case XK_Shift_R:
			keyports[0] &= 0xfe; break;
		case XK_Meta_L:
		case XK_Alt_L:
			keyports[7] &= 0xfd; break;
#ifdef DEC
		/* Use DEC "Compose Character" key for <edit> if available
		   else use RHS Alt/Meta key*/
		case XK_Multi_key: 
#endif
		case XK_Alt_R: case XK_Meta_R: case XK_Mode_switch:
		        keyports[0] &= 0xfe; 
			keyports[3] &= 0xfe; break;
		case XK_BackSpace: case XK_Delete:
		        keyports[0] &= 0xfe; 
		        keyports[4] &= 0xfe;
		        break;
		case XK_Escape: case XK_grave:
			keyports[3] &= 0xdf; break;
		case XK_Up:
		        keyports[0] &= 0xfe;
		        keyports[4] &= 0xf7; break;
		case XK_Down:
		        keyports[0] &= 0xfe;
		      	keyports[4] &= 0xef; break;
		case XK_Left:
		        keyports[0] &= 0xfe;
		        keyports[3] &= 0xef; break;
		case XK_Right:
		        keyports[0] &= 0xfe;
		        keyports[4] &= 0xfb; break;
		case XK_minus:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
                        keyports[6] &= 0xf7; break;
		case XK_underscore:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[4] &= 0xfe; break;
		case XK_equal:
			keyports[7] &= 0xfd; keyports[6] &= 0xfd;
			break;
		case XK_plus:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[6] &= 0xfb; break;
		case XK_semicolon:
			keyports[7] &= 0xfd; keyports[5] &= 0xfd;
			break;
		case XK_colon:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[0] &= 0xfd; 
			break;
		case XK_apostrophe:
			keyports[7] &= 0xfd; keyports[4] &= 0xf7; break;
		case XK_comma:
			keyports[7] &= 0xfd; keyports[7] &= 0xf7; break;
		case XK_period:
			keyports[7] &= 0xfd; keyports[7] &= 0xfb; break;
		case XK_less:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xf7; break;
		case XK_greater:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[2] &= 0xef; break;
		case XK_slash:
			keyports[7] &= 0xfd; keyports[0] &= 0xef; break;
		case XK_question:
			keyports[0] |= 0x01; keyports[7] &= 0xfd;
			keyports[0] &= 0xf7; break;
		case XK_backslash:
		        keyports[0] &= 0xfe; keyports[7] &= 0x7f;
			break;
		case XK_bar:
		        keyports[0] |= 0x01; keyports[7] &= 0xfd;
		        keyports[4] &= 0xfd;
			break;
                case XK_bracketleft:
                	keyports[7] &= 0xfd; keyports[2] &= 0xf7;
			break;
                case XK_bracketright:
                	keyports[7] &= 0xfd; keyports[2] &= 0xef;
			break;
                case XK_braceleft:
                        keyports[0] |= 0x01;
                        keyports[7] &= 0xfd; keyports[1] &= 0xfb;
                        break;                        
                case XK_braceright:
                        keyports[0] |= 0x01;
                        keyports[7] &= 0xfd; keyports[1] &= 0xfd;
                        break;                        
		case XK_quotedbl:
			keyports[0] |= 0x01;
			keyports[7] &= 0xfd; keyports[5] &= 0xfe;
			break;
		case XK_asciitilde:
			keyports[4] &= 0xfe;
			break;
		case XK_asciicircum:
		        keyports[0] |= 0x01; keyports[7] &= 0xfd;
		        keyports[6] &= 0xef;
		        break;
		/* hash and sterling characters must be treated 
		   separately as their position depends on the 
		   type of keyboard - UK or US */
		case XK_sterling:
                        keyports[0] |= 0x01;
                        keyports[7] &= 0xfd; keyports[6] &= 0xfd;
			break;
		case XK_numbersign:
			keyports[0] &= 0xfe; keyports[3] &=0xfb;
			break;

		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case XK_KP_0:
			keyports[5] &= 0x7f;
			break;
		case XK_KP_1:
			keyports[0] &= 0xdf;
			break;
		case XK_KP_2:
			keyports[0] &= 0xbf;
			break;
		case XK_KP_3:
			keyports[0] &= 0x7f;
			break;
		case XK_KP_4:
			keyports[1] &= 0xdf;
			break;
		case XK_KP_5:
			keyports[1] &= 0xbf;
			break;
		case XK_KP_6:
			keyports[1] &= 0x7f;
			break;
		case XK_KP_7:
			keyports[2] &= 0xdf;
			break;
		case XK_KP_8:
			keyports[2] &= 0xbf;
			break;
		case XK_KP_9:
			keyports[2] &= 0x7f;
			break;
                
                /* Finally the <inv> and (c) keys are mapped onto the
                   keypad <enter> and <.> */
                case XK_KP_Enter:
                        keyports[7] &= 0x7f;
                	break;
                case XK_KP_Decimal:
			keyports[0] &= 0xfe; keyports[5] &= 0xbf;
                	break;

		default:
			if (ks < 128) {
				keyports[keytable[ks].port] &= keytable[ks].bit;
			}
			break;
	}
}

/* Process a KeyRelease event */
void
process_spectrum_keyrelease(XKeyEvent *kev)
{
	char buf[3];
	KeySym ks;

	XLookupString(kev, buf, 2, &ks, NULL);

	if (ks == key_table[KM_JOY_UP])
		inPorts[0x1f] &= ~0x02;
	else if (ks == key_table[KM_JOY_DOWN])
		inPorts[0x1f] &= ~0x04;
	else if (ks == key_table[KM_JOY_LEFT])
		inPorts[0x1f] &= ~0x10;
	else if (ks == key_table[KM_JOY_RIGHT])
		inPorts[0x1f] &= ~0x08;
	else if (ks == key_table[KM_JOY_FIRE])
		inPorts[0x1f] &= ~0x01;

	switch (ks) {

		/* Ordinary SAM keys */
		case XK_Tab:
			keyports[3] |= ~0xbf; break;

		case XK_Return:
			keyports[6] |= ~0xfe; break;
		case XK_Control_L: case XK_Control_R:
		/* <Shift> <Control> is mapped to Caps Lock */
			     keyports[3] |= ~0xfd;
			break;
		case XK_Shift_L: case XK_Shift_R:
			keyports[0] |= ~0xfe; break;
		case XK_Meta_L:
		case XK_Alt_L: 
			keyports[7] |= ~0xfd; break;
#ifdef DEC
		/* Use DEC "Compose Character" key for <edit> if available
		   else use RHS Alt/Meta key*/
		case XK_Multi_key: 
#endif
		case XK_Alt_R: case XK_Meta_R: case XK_Mode_switch:
			keyports[0] |= ~0xfe;
			keyports[3] |= ~0xfe; break;
		case XK_BackSpace: case XK_Delete:
		        keyports[0] |= ~0xfe; 
		        keyports[4] |= ~0xfe;
		        break;
		case XK_Escape: case XK_grave:
			keyports[3] |= ~0xdf; break;
		case XK_Up:
		        keyports[0] |= ~0xfe;
		        keyports[4] |= ~0xf7; break;
		case XK_Down:
		        keyports[0] |= ~0xfe;
		      	keyports[4] |= ~0xef; break;
		case XK_Left:		
		        keyports[0] |= ~0xfe;
		        keyports[3] |= ~0xef; break;
		case XK_Right:
		        keyports[0] |= ~0xfe;
		        keyports[4] |= ~0xfb; break;
		case XK_minus:
			keyports[7] |= ~0xfd;
                        keyports[6] |= ~0xf7; break;
		case XK_underscore:
			keyports[7] |= ~0xfd;
			keyports[4] |= ~0xfe; break;
		case XK_equal:
			keyports[7] |= ~0xfd; keyports[6] |= ~0xfd; break;
		case XK_plus:
			keyports[7] |= ~0xfd;
			keyports[6] |= ~0xfb; break;
		case XK_semicolon:
			keyports[7] |= ~0xfd; keyports[5] |= ~0xfd; break;
		case XK_colon:
			keyports[7] |= ~0xfd;
			keyports[0] |= ~0xfd; break;
		case XK_apostrophe:
			keyports[7] |= ~0xfd; keyports[4] |= ~0xf7; break;
		case XK_comma:
			keyports[7] |= ~0xfd; keyports[7] |= ~0xf7; break;
		case XK_period:
			keyports[7] |= ~0xfd; keyports[7] |= ~0xfb; break;
		case XK_less:
			keyports[7] |= ~0xfd;
			keyports[2] |= ~0xf7; break;
		case XK_greater:
			keyports[7] |= ~0xfd;
			keyports[2] |= ~0xef; break;
		case XK_slash:
			keyports[7] |= ~0xfd; keyports[0] |= ~0xef; break;
		case XK_question:
			keyports[7] |= ~0xfd;
			keyports[0] |= ~0xf7; break;
		case XK_backslash:
		        keyports[0] |= ~0xfe; keyports[7] |= ~0x7f;
			break;
		case XK_bar:
		        keyports[7] |= ~0xfd;
		        keyports[4] |= ~0xfd;
			break;
                case XK_bracketleft:
                	keyports[7] |= ~0xfd; keyports[2] |= ~0xf7;
			break;
                case XK_bracketright:
                	keyports[7] |= ~0xfd; keyports[2] |= ~0xef;
			break;
                case XK_braceleft:
                        keyports[7] |= ~0xfd; keyports[1] |= ~0xf7;
                        break;                        
                case XK_braceright:
                        keyports[7] |= ~0xfd; keyports[1] |= ~0xfd;
                        break;                        
		case XK_quotedbl:
			keyports[7] |= ~0xfd; keyports[5] |= ~0xfe; 
			break;
		case XK_asciitilde:
			keyports[4] |= ~0xfe;
			break;
		case XK_asciicircum:
		        keyports[7] |= ~0xfd;
		        keyports[6] |= ~0xef;
			break;
		/* hash and sterling characters must be treated 
		   separately as their position depends on the 
		   type of keyboard - UK or US */
		case XK_sterling:
                        keyports[7] |= ~0xfd; keyports[6] |= ~0xfd;
			break;
		case XK_numbersign:
			keyports[0] |= ~0xfe; keyports[3] |= ~0xfb;
			break;


		/* The sam function keys are mapped onto the numeric
		   keypad */
		   
		case XK_KP_0:
			keyports[5] |= ~0x7f;
			break;
		case XK_KP_1:
			keyports[0] |= ~0xdf;
			break;
		case XK_KP_2:
			keyports[0] |= ~0xbf;
			break;
		case XK_KP_3:
			keyports[0] |= ~0x7f;
			break;
		case XK_KP_4:
			keyports[1] |= ~0xdf;
			break;
		case XK_KP_5:
			keyports[1] |= ~0xbf;
			break;
		case XK_KP_6:
			keyports[1] |= ~0x7f;
			break;
		case XK_KP_7:
			keyports[2] |= ~0xdf;
			break;
		case XK_KP_8:
			keyports[2] |= ~0xbf;
			break;
		case XK_KP_9:
			keyports[2] |= ~0x7f;
			break;
                
                /* Finally the <inv> and (c) keys are mapped onto the
                   keypad <enter> and <.> */
                case XK_KP_Enter:
                        keyports[7] |= ~0x7f;
                	break;
                case XK_KP_Decimal:
			keyports[0] |= ~0xfe; keyports[5] |= ~0xbf;
                	break;

		default:
			if (ks < 128)
				keyports[keytable[ks].port] |= ~keytable[ks].bit;
			break;
	}
}
